@@ -1,7 +1,10 @@
 from __future__ import annotations
 
 import contextlib
-from typing import Any, Literal, TypeVar
+import copy
+import gc
+import json
+from typing import Any, Callable, Literal, TypeVar
 
 import torch
 import torch.nn.functional as F  # noqa: N812
@@ -9,9 +12,11 @@ import transformers
 from loguru import logger
 from transformers import AutoModelForCausalLM, AutoTokenizer, BatchEncoding, PreTrainedModel, PreTrainedTokenizer
 
+from flexeval.core.string_processor import StringProcessor
+from flexeval.core.tool_parser.base import ToolParser
 from flexeval.utils.hf_utils import get_default_model_kwargs
 
-from .base import LanguageModel, normalize_stop_sequences
+from .base import LanguageModel, LMOutput, normalize_stop_sequences
 
 T = TypeVar("T")
 
@@ -56,7 +61,7 @@ def tokenize_text_for_lm_prefix(
 def tokenize_text_for_lm_continuation(
     text_list: list[str],
     tokenizer: PreTrainedTokenizer,
-    oov_character: str = "彁",
+    single_token_character: str = "\x80",  # noqa: S107
     as_continuation: bool | list[bool] = True,
 ) -> BatchEncoding:
     """When tokenizing a prefix and continuation separately, the sentencepiece-
@@ -82,17 +87,14 @@ def tokenize_text_for_lm_continuation(
         msg = "The length of as_continuation must be the same as the length of text_list."
         raise ValueError(msg)
 
-    if oov_character in tokenizer.get_vocab():
-        msg = f"oov_character '{oov_character}' is already in the tokenizer's vocab."
-        raise ValueError(msg)
-    oov_char_len = len(tokenizer.tokenize(oov_character))
+    oov_char_len = len(tokenizer.tokenize(single_token_character))
 
     encoding_list: list[BatchEncoding] = []
     for text, as_cont in zip(text_list, as_continuation):
         input_text = text
         # tokenize with OOV character
         if as_cont:
-            input_text = oov_character + text
+            input_text = single_token_character + text
         encoding = tokenizer(
             input_text,
             add_special_tokens=False,
@@ -108,6 +110,56 @@ def tokenize_text_for_lm_continuation(
         return tokenizer.pad(encoding_list, return_tensors="pt")
 
 
+def decode_for_lm_continuation(
+    output_tokens: list[int],
+    input_tokens: list[int],
+    tokenizer: PreTrainedTokenizer,
+) -> str:
+    """
+    Decode output tokens while preserving correct formatting.
+
+    Some tokenizers modify tokenized text (e.g., stripping leading spaces). This function ensures
+    the continuation text is extracted correctly by decoding both input and output tokens together
+    and removing the input portion.
+
+    Example of undesired behavior:
+    ```
+    # ['▁▁▁', '▁return', '▁1']
+    >>> tokenizer.decode([1, 2, 3])
+    '   return 1'  # The first space is stripped away due to the tokenizer's behavior.
+    ```
+
+    Example of desired behavior:
+    ```
+    # input_tokens: ['▁def', '▁func', '():', '\n']
+    # output_tokens: ['▁▁▁', '▁return', '▁1']
+    >>> decode_for_lm_continuation(output_tokens = [1, 2, 3], input_tokens = [1])
+    '    return 1'  # The first space is preserved.
+    """
+    entire_text = tokenizer.decode(input_tokens + output_tokens, skip_special_tokens=False)
+    input_text = tokenizer.decode(input_tokens, skip_special_tokens=False)
+    return entire_text[len(input_text) :]
+
+
+def deserialize_tool_calls_in_messages(messages: list[dict[str, Any]]) -> None:
+    """
+    We adopt the standard OpenAI format, where the 'arguments' field in tool_calls is expected to be a JSON string.
+    However, huggingface/transformers expects the 'arguments' field to be a dict.
+    https://huggingface.co/docs/transformers/v4.48.2/chat_templating#a-complete-tool-use-example
+
+    To resolve this mismatch, this function deserializes 'arguments' before passing messages to apply_chat_template.
+
+    Args:
+        messages: A list of messages to deserialize.
+    """
+    deserialized_messages = copy.deepcopy(messages)
+    for message in deserialized_messages:
+        if message["role"] == "assistant" and "tool_calls" in message:
+            for item in message["tool_calls"]:
+                item["function"]["arguments"] = json.loads(item["function"]["arguments"])
+    return deserialized_messages
+
+
 class HuggingFaceLM(LanguageModel):
     """
     LanguageModel implementation using Hugging Face Transformers.
@@ -124,6 +176,16 @@ class HuggingFaceLM(LanguageModel):
         load_peft: Should be set to True when loading the model from PEFT weights.
         custom_chat_template: A custom chat template for chatbot models.
             If specified, this overrides the default chat template of the tokenizer.
+        system_message: System messages to be prepended to given messages. It applies only for
+            chat response.
+        default_gen_kwargs: Default generation kwargs to use when calling the API.
+        string_processors: A single or a list of StringProcessor objects to process the model's output.
+        model_limit_tokens: An upper limit on the number of tokens (input + output) the model can handle.
+            If `max_new_tokens` exceeds this limit in `generate_chat_response()`, it will be capped to this value.
+            If this value is set to less than or equal to the model's capacity and the input exceeds it,
+            an empty string is returned instead of raising an error.
+            If set to “default”, the value will be automatically determined when possible.
+        tool_parser: A ToolParser object to extract the tool_calls from the model's output.
     """
 
     def __init__(
@@ -137,40 +199,73 @@ class HuggingFaceLM(LanguageModel):
         random_seed: int = 42,
         load_peft: bool = False,
         custom_chat_template: str | None = None,
+        chat_template_kwargs: dict[str, Any] | None = None,
+        system_message: str | None = None,
         default_gen_kwargs: dict[str, Any] | None = None,
+        string_processors: StringProcessor | list[StringProcessor] | None = None,
+        model_limit_tokens: int | None | Literal["default"] = "default",
+        tool_parser: ToolParser | None = None,
     ) -> None:
+        super().__init__(string_processors=string_processors)
         self._model_name_or_path = model
         tokenizer = tokenizer if tokenizer else model
         tokenizer_kwargs = tokenizer_kwargs or {}
         self.tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(tokenizer, **tokenizer_kwargs)
         self.custom_chat_template = custom_chat_template
+        self.system_message = system_message
+        self.chat_template_kwargs = chat_template_kwargs or {}
         self.add_special_tokens = add_special_tokens
         self.default_gen_kwargs = default_gen_kwargs or {}
-
-        model_kwargs = get_default_model_kwargs(model_kwargs)
-        if not load_peft:
-            self.model: PreTrainedModel = AutoModelForCausalLM.from_pretrained(
-                model,
-                **model_kwargs,
-            )
-        else:
-            from peft import AutoPeftModelForCausalLM
-
-            self.model = AutoPeftModelForCausalLM.from_pretrained(
-                model,
-                **model_kwargs,
-            )
-
-        self.model.eval()
-
+        # `self.model` is initialized lazily to avoid unnecessary memory usage.
+        self.model: PreTrainedModel | None = None
+        self.model_kwargs = get_default_model_kwargs(model_kwargs)
+        self.load_peft = load_peft
         self.amp_dtype = amp_dtype
-
-        transformers.set_seed(random_seed)
-
-        logger.info(f"model device: {self.model.device}")
-        logger.info(f"model dtype: {self.model.dtype}")
+        self.model_limit_tokens = model_limit_tokens
+        self.tool_parser = tool_parser
         logger.info(f"amp_dtype: {amp_dtype}")
         logger.info(f"random seed: {random_seed}")
+        transformers.set_seed(random_seed)
+
+    @staticmethod
+    def load_model(method: Callable) -> Callable:
+        """Decorator to load the model lazily."""
+
+        def wrapper(self: HuggingFaceLM, *args: tuple, **kwargs: dict) -> Callable:
+            if self.model is None:
+                if not self.load_peft:
+                    self.model = AutoModelForCausalLM.from_pretrained(
+                        self._model_name_or_path,
+                        **self.model_kwargs,
+                    )
+                else:
+                    from peft import AutoPeftModelForCausalLM
+
+                    self.model = AutoPeftModelForCausalLM.from_pretrained(
+                        self._model_name_or_path,
+                        **self.model_kwargs,
+                    )
+
+                self.model.eval()
+
+                if self.model_limit_tokens == "default":
+                    hf_config = self.model.config.to_dict()
+                    if "n_positions" in hf_config:
+                        self.model_limit_tokens = hf_config["n_positions"]
+                    elif "max_position_embeddings" in hf_config:
+                        self.model_limit_tokens = hf_config["max_position_embeddings"]
+                    else:
+                        msg = (
+                            "`model_limit_tokens` was set to “default”, but the default max_position_embedeings "
+                            "could not be found in the config. Set it to `None`."
+                        )
+                        logger.warning(msg)
+
+                logger.info(f"model device: {self.model.device}")
+                logger.info(f"model dtype: {self.model.dtype}")
+            return method(self, *args, **kwargs)
+
+        return wrapper
 
     def _get_amp_context(self) -> contextlib.AbstractContextManager:
         if self.amp_dtype is None:
@@ -214,15 +309,15 @@ class HuggingFaceLM(LanguageModel):
         return stop_token_ids
 
     @torch.inference_mode()
-    def batch_complete_text(
+    @load_model
+    def _batch_complete_text(
         self,
         text_list: list[str],
         stop_sequences: str | list[str] | None = None,
         max_new_tokens: int | None = None,
         ignore_eos: bool = False,
-        include_stop_str_in_output: bool = False,
         **kwargs,
-    ) -> list[str]:
+    ) -> list[LMOutput]:
         gen_kwargs = self.default_gen_kwargs.copy()
         gen_kwargs.update(kwargs)
         if max_new_tokens is not None:
@@ -235,6 +330,19 @@ class HuggingFaceLM(LanguageModel):
         ).to(self.model.device)
         input_token_length = model_inputs["input_ids"].shape[1]
 
+        if self.model_limit_tokens:
+            model_limit_new_tokens = self.model_limit_tokens - input_token_length
+            if model_limit_new_tokens <= 0:
+                msg = (
+                    f"Received input that is longer than `model_limit_tokens = {self.model_limit_tokens}`. "
+                    f"This batch returns empty strings."
+                )
+                logger.warning(msg)
+                return [LMOutput(text="", finish_reason="input_length_limit") for _ in text_list]
+
+            if "max_new_tokens" not in gen_kwargs or model_limit_new_tokens < gen_kwargs["max_new_tokens"]:
+                gen_kwargs["max_new_tokens"] = model_limit_new_tokens
+
         # set the stop sequences
         stop_sequences = normalize_stop_sequences(
             stop_sequences_list=[
@@ -242,6 +350,7 @@ class HuggingFaceLM(LanguageModel):
                 gen_kwargs.pop("stop_strings", None),  # This is used in the transformers `generate` function
                 gen_kwargs.pop("stop_sequences", None),  # This is a common variable name used in flexeval
             ],
+            bos_token=self.tokenizer.bos_token,
             eos_token=self.tokenizer.eos_token,
             ignore_eos=ignore_eos,
         )
@@ -254,44 +363,67 @@ class HuggingFaceLM(LanguageModel):
         )
 
         with self._get_amp_context():
-            lm_outputs = self.model.generate(**model_inputs, **gen_kwargs)
+            generated_tokens = self.model.generate(**model_inputs, **gen_kwargs)
 
-        # `lm_outputs` contains full text including the input text.
         # We strip the input text and stop sequences from the output text.
-        output_texts: list[str] = []
-        for output_tensor in lm_outputs[:, input_token_length:]:
+        lm_outputs: list[LMOutput] = []
+        for generated_tensor in generated_tokens:
+            input_tensor = generated_tensor[:input_token_length]
+            output_tensor = generated_tensor[input_token_length:]
+
+            input_tokens = [t for t in input_tensor.tolist() if t != self.tokenizer.pad_token_id]
             output_tokens = [t for t in output_tensor.tolist() if t != self.tokenizer.pad_token_id]
-            decoded_text = self.tokenizer.decode(output_tokens, skip_special_tokens=False)
-
-            if include_stop_str_in_output:
-                output_texts.append(decoded_text)
-                continue
+            decoded_text = decode_for_lm_continuation(output_tokens, input_tokens, self.tokenizer)
 
+            finish_reason = "length"
             for stop_seq in stop_sequences:
                 idx = decoded_text.find(stop_seq)
                 if idx != -1:
                     decoded_text = decoded_text[:idx]
-            output_texts.append(decoded_text)
-        return output_texts
+                    finish_reason = "stop"
+            lm_outputs.append(LMOutput(text=decoded_text, finish_reason=finish_reason))
+        return lm_outputs
 
-    def batch_generate_chat_response(
+    @load_model
+    def _batch_generate_chat_response(
         self,
-        chat_messages_list: list[list[dict[str, str]]],
+        chat_messages_list: list[list[dict[str, Any]]],
+        tools_list: list[list[dict[str, Any]] | None] | None = None,
         **kwargs,
-    ) -> list[str]:
+    ) -> list[LMOutput]:
+        if tools_list is None:
+            tools_list = [None] * len(chat_messages_list)
+        if self.system_message is not None:
+            for chat_messages in chat_messages_list:
+                chat_messages.insert(0, {"role": "system", "content": self.system_message})
         chat_messages_as_string = [
             self.tokenizer.apply_chat_template(
-                chat_messages,
+                deserialize_tool_calls_in_messages(chat_messages),
+                tools=tools,
                 tokenize=False,
                 add_generation_prompt=True,
                 chat_template=self.custom_chat_template,
+                **self.chat_template_kwargs,
             )
-            for chat_messages in chat_messages_list
+            for chat_messages, tools in zip(chat_messages_list, tools_list)
         ]
-        return self.batch_complete_text(chat_messages_as_string, **kwargs)
+        lm_outputs = self._batch_complete_text(chat_messages_as_string, **kwargs)
+        if self.tool_parser:
+            for lm_output, tools in zip(lm_outputs, tools_list):
+                if tools is None:
+                    continue
+                lm_output: LMOutput
+                parsed_tool_calling_message = self.tool_parser(lm_output.text)
+                lm_output.tool_calls = parsed_tool_calling_message.tool_call_dicts
+                lm_output.raw_text = parsed_tool_calling_message.raw_text
+                lm_output.text = parsed_tool_calling_message.text
+                lm_output.tool_call_validation_result = parsed_tool_calling_message.validation_result
+
+        return lm_outputs
 
     @torch.inference_mode()
-    def batch_compute_log_probs(
+    @load_model
+    def _batch_compute_log_probs(
         self,
         text_list: list[str],
         prefix_list: list[str] | None = None,
@@ -397,8 +529,9 @@ class HuggingFaceLM(LanguageModel):
             total_log_probs = (log_prob_of_next * log_prob_mask).sum(dim=-1)
         return total_log_probs.tolist()
 
-    def batch_compute_chat_log_probs(
-        self, prompt_list: list[list[dict[str, str]]], response_list: list[dict[str, str]]
+    @load_model
+    def _batch_compute_chat_log_probs(
+        self, prompt_list: list[list[dict[str, Any]]], response_list: list[dict[str, Any]]
     ) -> list[float]:
         prompt_as_string: list[str] = []
         response_as_string: list[str] = []
@@ -411,7 +544,15 @@ class HuggingFaceLM(LanguageModel):
             )
             prompt_as_string.append(prompt_as_string_i)
             response_as_string.append(response_as_string_i)
-        return self.batch_compute_log_probs(prompt_as_string, prefix_list=response_as_string)
+        return self._batch_compute_log_probs(response_as_string, prefix_list=prompt_as_string)
+
+    def cleanup_resources(self) -> None:
+        del self.model
+        self.model = None
+        gc.collect()
+        if torch.cuda.is_available():
+            logger.info("Cleaning up CUDA resources...")
+            torch.cuda.empty_cache()
 
     def __repr__(self) -> str:
         return f"{self.__class__.__name__}(model={self._model_name_or_path!r})"
